from __future__ import annotations
import json
from pathlib import Path
from typing import Callable, Literal, TypeVar, overload, Any
from ._json_class import JsonClass, _JSON_TEMPLATE, _JSON_MUTABLE
from ._json_attribute import JsonProperty

_C = TypeVar("_C")

@overload
def jsonclass(template_or_class: type[_C], template: dict[str, Any | None], mutable: bool = False) -> type[_C | JsonClass]:
    ...

@overload
def jsonclass(template_or_class: Literal[None], template: dict[str, Any | None], mutable: bool = False) -> Callable[[type[_C]], type[_C | JsonClass]]:
    ...
    
@overload
def jsonclass(template_or_class: dict[str, Any | None], template: Literal[None] = None, mutable: bool = False) -> Callable[[type[_C]], type[_C | JsonClass]]:
    ...

    
def jsonclass(template_or_class=None, template=None, mutable=False):
    """
    Create a json class with specified template.
    
    This decorator is an alternative of ``JsonClass`` inheritance. Generally this
    decorator will be used as following.
    
    >>> @jsonclass({...})
    >>> class C: ...
    
    >>> @jsonclass
    >>> class C: 
    >>>     __json_template__ = {...}

    Returns
    -------
    subclass of JsonClass
        An class that inherits the input class and ``JsonClass``.
    
    Excamples
    ---------
    >>> template = {..., {..., "xxx": "a"}}
    >>> @jsonclass(template)
    >>> class C:
    >>>     a: int
    >>> c = C()
    >>> c.a  # get some value
    """
    if isinstance(template_or_class, type):
        cls = template_or_class
    elif isinstance(template_or_class, dict):
        cls = None
        template = template_or_class
    elif template_or_class is None:
        cls = None
    else:
        raise TypeError
    
    def _func(cls_):
        nonlocal template, mutable
        template = getattr(cls, _JSON_TEMPLATE, template)
        mutable = getattr(cls, _JSON_MUTABLE, mutable)
        if not isinstance(template, dict):
            raise TypeError("`template` must be given as a dict.")
        ns = {_JSON_TEMPLATE: template, _JSON_MUTABLE: mutable}
        return type(cls_.__name__, (cls_, JsonClass), ns)
    
    return _func if cls is None else _func(cls)

class _dummy:
    """Dummy class for json class creation."""

def create_constructor(
    template: dict[str, Any | None],
    mutable: bool = False,
    name: str | None = None
) -> type[_dummy | JsonClass]:
    """
    Create a JsonClass in a simple way.
    
    Instead of defining a base class, this function uses a dummy class and create
    a constructor using a template dictionary only. As a result, typing of json
    properties will be lost.
    
    Parameters
    ----------
    template : dict
        JSON file template dictionary.
    mutable : bool, default is False
        Default mutability of properties.
    name : str, optional
        Name of the class. Automatically determined by default.
    
    Returns
    -------
    JsonClass subclass
        A class implemented with properties extracted from the template dictionary.
    
    See also
    --------
    :func:`create_loader`
    """
    cls = jsonclass(_dummy, template=template, mutable=mutable)
    if name is None:
        cls.__name__ = f"JsonClass{hex(id(cls))}"
    else:
        cls.__name__ = str(name)
    cls.__qualname__ = f"elegant_json.{cls.__name__}"
    cls.__module__ = "elegant_json"
    return cls

def create_loader(
    template: dict[str, Any | None],
    mutable: bool = False,
    name: str | None = None
):
    """
    Create a loader function in a simple way.
    
    Unlike :func:`create_constructor`, this function returns a loader function
    that takes file path as an input and returns a JsonClass object.
    
    Parameters
    ----------
    template : dict
        JSON file template dictionary.
    mutable : bool, default is False
        Default mutability of properties.
    name : str, optional
        Name of the class. Automatically determined by default.
    
    Returns
    -------
    Callable
        A loader function.
    
    See also
    --------
    :func:`create_constructor`
    """
    cls = create_constructor(template=template, mutable=mutable, name=name)
    # NOTE: simply this function can return `cls.load` but will not work if
    # new class has `load` property by chance.
    def load(path: str | Path | bytes, encoding: str | None = None):
        with open(path, mode="r", encoding=encoding) as f:
            js = json.load(f)
        return cls(js)  # type: ignore
    return load


def isformatted(obj, json_class: type[JsonClass]) -> bool:
    """
    Check if the input object is in ``json_class`` format.
    
    This function is free of error, as long as a ``JsonClass`` object is given as
    the second argument.
    """
    if not issubclass(json_class, JsonClass):
        raise TypeError("The second argument of `isformatted` must be a JsonClass.")
    if not isinstance(obj, dict):
        return False
    for name in json_class._json_properties:
        prop: JsonProperty = getattr(json_class, name)
        try:
            out: Any = obj
            for k in prop.keys():
                out = out[k]
        except (KeyError, IndexError):
            return False
    return True


from __future__ import annotations
import json
from pathlib import Path
from typing import Any, Iterable

from ._json_attribute import Attr

_JSON_TEMPLATE = "__json_template__"
_JSON_MUTABLE = "__json_mutable__"

def _iter_dict(d: dict[str, Any | None], keys: list[str | int]) -> Iterable[tuple[Any | None, list[str | int]]]:
    for k, v in d.items():
        next_keys = keys + [k]
        if isinstance(v, (list, tuple)):
            yield from  _iter_list(v, next_keys)
        elif isinstance(v, dict):
            yield from  _iter_dict(v, next_keys)
        else:
            yield v, next_keys

def _iter_list(l: Iterable[Any | None], keys: list[str | int]) -> Iterable[tuple[Any | None, list[str | int]]]:
    for k, v in enumerate(l):
        next_keys = keys + [k]
        if isinstance(v, (list, tuple)):
            yield from  _iter_list(v, next_keys)
        elif isinstance(v, dict):
            yield from  _iter_dict(v, next_keys)
        else:
            yield v, next_keys


class JsonClassMeta(type):
    """The metaclass of JsonClass."""
    
    __json_template__: dict[str, Any | None] = {}
    __json_mutable__: bool = False
    _json_properties: frozenset[str]

    def __new__(
        cls: type,
        name: str,
        bases: tuple[type, ...],
        namespace: dict[str, Any],
        **kwds,
    ) -> JsonClassMeta:
        
        _js_temp = namespace.get(_JSON_TEMPLATE, {})
        _mutable = namespace.get(_JSON_MUTABLE, False)
        _annot = namespace.get("__annotations__", {})
        props = set()
        
        for attr, keys in _iter_dict(_js_temp, []):
            if isinstance(attr, Attr):
                if attr.name is None:
                    attr_name = keys[-1]
                    if not isinstance(attr_name, str):
                        raise TypeError(
                            "Attr objects in a list need `name` argument."
                        )
                    attr.name = attr_name
                if not attr.mutability_given:
                    attr.mutable = _mutable
                
                if attr.annotation is None:
                    attr.annotation = _annot.get(attr.name)
                
            else:
                continue
            
            # check name collision
            if attr.name in props:
                raise ValueError(f"Name collision in attributes: {attr.name!r}.")
            props.add(attr.name)
            
            # convert into a json-property
            prop = attr.to_property(keys)
            namespace[attr.name] = prop
        
        jcls: JsonClassMeta = type.__new__(cls, name, bases, namespace, **kwds)
        jcls._json_properties = frozenset(props)
        
        return jcls


class JsonClass(metaclass=JsonClassMeta):
    """The base class of json class."""
    
    def __init__(self, d: dict[str, Any | None], /):
        if not isinstance(d, dict):
            raise TypeError(
                f"Input of {self.__class__.__name__} must be a dict, got {type(d)}"
            )
        self._json = d
    
    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} object>"
    
    @property
    def json(self) -> dict[str, Any | None]:
        """Return the original json dictionary."""
        return self._json
    
    @classmethod
    def load(cls, path: str | Path | bytes, encoding: str | None = None):
        """Load a json file and create a json class from it."""
        with open(path, mode="r", encoding=encoding) as f:
            js = json.load(f)
        return cls(js)
    
    @classmethod
    def loads(cls, s: str | bytes):
        """Deserialize input string and create a json class from it."""
        js: dict[str, Any | None] = json.loads(s)
        return cls(js)

    def dump(self, path: str | Path | bytes, encoding: str | None = None) -> None:
        """Save json object in a file."""
        with open(path, mode="w", encoding=encoding) as f:
            json.dump(self.json, f)
        return None

    def attr_asdict(self) -> dict[str, Any | None]:
        """Summarize JsonClass properties into a dict."""
        return {
            name: getattr(self, name)
            for name in self.__class__._json_properties
        }

    def attr_astuple(self) -> tuple[Any | None, ...]:
        """Summarize JsonClass properties into a tuple."""
        return tuple(
            getattr(self, jprop_name) 
            for jprop_name in self.__class__._json_properties
        )


__version__ = "0.1.0"

from .core import (
    jsonclass,
    create_loader,
    create_constructor,
    isformatted,
)

from ._json_class import JsonClass, Attr

__all__ = [
    "Attr",
    "JsonClass",
    "jsonclass",
    "create_loader",
    "create_constructor",
    "isformatted",
]


from setuptools import setup, find_packages

with open("elegant_json/__init__.py", encoding="utf-8") as f:
    line = next(iter(f))
    VERSION = line.strip().split()[-1][1:-1]

with open("README.md") as f:
    readme = f.read()

setup(
    name="elegant-json",
    version=VERSION,
    description="Deal with JSON in an elegant way",
    long_description=readme,
    long_description_content_type="text/markdown",
    author="Hanjin Liu",
    author_email="liuhanjin-sc@g.ecc.u-tokyo.ac.jp",
    license="BSD 3-Clause",
    download_url="https://github.com/hanjinliu/elegant-json",
    packages=find_packages(exclude=["docs", "examples", "rst", "tests", "tests.*"]),
    package_data={"elegant_json": ["**/*.pyi", "*.pyi"]},
    install_requires=[],
    python_requires=">=3.8",
)



from __future__ import annotations
from types import GenericAlias
from typing import Any, TYPE_CHECKING, ForwardRef, get_args, get_origin
from typing import _eval_type  # type: ignore

if TYPE_CHECKING:
    from ._json_class import JsonClass


def _define_converter(annotation: type | GenericAlias | ForwardRef | None):
    if annotation is None:
        converter = lambda x: x
    elif isinstance(annotation, GenericAlias):
        origin = get_origin(annotation)
        args = get_args(annotation)
        if origin is list:
            arg = args[0]
            if isinstance(arg, type):
                converter = lambda x: list(_define_converter(arg)(a) for a in x)
            else:
                raise ValueError(f"Wrong type annotation: {annotation!r}.")
        elif origin is dict:
            if args[0] is not str:
                raise TypeError("Only dict[str, ...] is supported.")
            val = args[1]
            if isinstance(val, type):
                converter = lambda x: {k: _define_converter(val)(v) for k, v in x.items()}
            else:
                raise ValueError(f"Wrong type annotation: {annotation!r}.")
        elif origin is tuple:
            converter = lambda x: tuple(_define_converter(arg)(a) for arg, a in zip(args, x))
        else:
            raise ValueError(f"Wrong type annotation: {annotation!r}.")
    elif isinstance(annotation, (str, ForwardRef)):
        if isinstance(annotation, str):
            annotation = ForwardRef(annotation)
        converter = _define_converter(_eval_type(annotation, None, None))
    else:
        converter = lambda x: annotation(x)
    return converter

class JsonProperty(property):
    def keys(self) -> list[str | int]:
        return self._keys
    
    def set_keys(self, keys):
        self._keys = keys
    
    def getter(self, fget) -> JsonProperty:
        return self.__class__(fget, self.fset, self.fdel, self.__doc__)
    
    def setter(self, fset) -> JsonProperty:
        return self.__class__(self.fget, fset, self.fdel, self.__doc__)


class Attr:
    """Attribute class for JsonClass."""
    
    def __init__(
        self,
        name: str | None = None, 
        annotation: type | None = None,
        *,
        default = None,
        mutable: bool | None = None,
    ):
        self.name = name
        self.default = default
        self.mutable = mutable or False
        self.mutability_given = mutable is not None
        self.annotation = annotation
        
    @property
    def name(self) -> str | None:
        return self._name
    
    @name.setter
    def name(self, value: str | None):
        if value is not None and not value.isidentifier():
            raise ValueError(f"{value!r} is not an identifier.")
        self._name = value
    
    def to_property(self, keys: list[str | int]) -> JsonProperty:
        converter = _define_converter(self.annotation)
        def fget(jself: JsonClass):
            out: Any = jself._json
            try:
                for k in keys:
                    out = out[k]
            except (KeyError, IndexError):
                out = self.default
            else:
                out = converter(out)
            return out
        
        prop = JsonProperty(fget)
        
        if self.mutable:
            def fset(jself: JsonClass, value):
                out: Any = jself._json
                for k in keys[:-1]:
                    out = out[k]
                out[keys[-1]] = value
                return None
        
            prop = prop.setter(fset)
        
        prop.set_keys(keys)
        return prop
